#PALM_UNIT   -name unit_Left\
#            -functions {python unit_Left}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_SPACE -name double_array\
#           -shape (:)\
#           -element_size PL_DOUBLE_PRECISION\
#           -comment {CGNS/Python double array}
#
#PALM_OBJECT -name Onode\
#            -space double_array\
#            -intent OUT\
#            -comment {CGNS/Python double array out}
#
#PALM_OBJECT -name SYNC\
#            -space one_integer\
#            -intent OUT\
#            -comment {CGNS/Python sync integer out}
#
import palm
import T
from C3PO import capsule

palm.init('NONE')
palm.write('CAPSULE CGNS/Python')
path='/CGNSTree/Cylindre/cyl0/GridCoordinates/CoordinateX'
capsule.nodePublishAsDoubleArray("Onode",'double_array.unit_Left',1,T.data,path)
palm.write('Publish %s'%path)
palm.finalize()

# --- last line




# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
import CGNS.PAT.cgnskeywords  as CGK
import CGNS.PAT.cgnsutils     as CGU
import CGNS.PAT.cgnslib       as CGL

import pickle
import numpy
import palm

def syncEmit(localname):
  nd=numpy.ones((1,),dtype='i')*92322
  po=palm.PalmObject(object=localname,space='one_integer')
  po.put(nd)
  return nd[0]
  
def syncWait(localname):
  po=palm.PalmObject(object=localname,space='one_integer')
  po.object_get_spacename()
  po.space_get_rank()
  po.space_get_shape()
  dt=numpy.empty(po.shape,dtype='i')
  po.get(dt)
  return dt[0]

def nodePublish(localname,space,rank,tree,path):
  nd=CGU.getNodeByPath(tree,path)
  treePublish(localname,space,rank,nd)
  
def nodeRetrieve(localname):
  return treeRetrieve(localname)

def nodePublishAsInteger(localname,rank,tree,path):
  nd=CGU.getNodeByPath(tree,path)
  po=palm.PalmObject(object=localname,space='one_integer')
  po.put(nd[1])
  
def nodePublishAsDouble(localname,rank,tree,path):
  nd=CGU.getNodeByPath(tree,path)
  po=palm.PalmObject(object=localname,space='one_double')
  po.put(nd[1])
  
def nodePublishAsDoubleArray(localname,space,rank,tree,path):
  nd=CGU.getNodeByPath(tree,path)
  po=palm.PalmObject(object=localname,space=space)
  po.shape=[nd[1].size]
  po.space=space
  po.rank=rank
  po.space_set_shape()
  f=syncEmit('SYNC.unit_Left')
  dd=numpy.array(nd[1].ravel(),dtype='d')
  po.put(dd)
  
def nodeRetrieveAsInteger(localname,tree=None,path=None):
  po=palm.PalmObject(object=localname,space='one_integer')
  po.object_get_spacename()
  po.space_get_rank()
  po.space_get_shape()
  dt=numpy.empty(po.shape,dtype='i')
  po.get(dt)
  if ((tree is None) or (path is None)):
    return dt[0]
  else:
    nd=CGU.getNodeByPath(tree,path)
    nd[1]=dt
 
def nodeRetrieveAsDouble(localname,tree=None,path=None):
  po=palm.PalmObject(object=localname,space='one_double')
  po.object_get_spacename()
  po.space_get_rank()
  po.space_get_shape()
  dt=numpy.empty(po.shape,dtype='d')
  po.get(dt)
  if ((tree is None) or (path is None)):
    return dt[0]
  else:
    nd=CGU.getNodeByPath(tree,path)
    nd[1]=dt
    

def nodeRetrieveAsDoubleArray(localname):
  po=palm.PalmObject(object=localname,space='NULL')
  po.object_get_spacename()
  po.space_get_rank()
  f=syncWait('SYNC.unit_Right')
  po.space_get_shape()
  dt=numpy.empty(po.shape,dtype='d')
  po.get(dt)
  return dt

def treePublish(localname,space,rank,tree):
  po=palm.PalmObject(object=localname,space=space)
  st=pickle.dumps(tree)
  padding=(4-len(st)%4)
  dt=numpy.array(st+' '*padding)
  po.shape=[len(st)+padding]
  po.space=space
  po.rank=rank
  po.space_set_shape()
  po.put(dt)

def treeRetrieve(localname):
  po=palm.PalmObject(object=localname,space="NULL")
  po.object_get_spacename()
  po.space_get_rank()
  po.space_get_shape()
  dt=numpy.empty(po.shape,dtype='c')
  po.get(dt)
  st=pickle.loads(dt.tostring().rstrip())
  return st

# --- last line



#PALM_UNIT   -name elsA_palm\
#            -functions {python elsA_palm}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_SPACE -name CGNSPYTHON\
#           -shape (:)\
#           -element_size PL_CHARACTER\
#           -comment {CGNS/Python sub-tree serialization}
#
#PALM_SPACE -name double_array\
#           -shape (:)\
#           -element_size PL_DOUBLE_PRECISION\
#           -comment {CGNS/Python double array}
#
#PALM_OBJECT -name Tnode\
#            -space CGNSPYTHON\
#            -intent OUT\
#            -comment {CGNS/Python node data out}
#
#PALM_OBJECT -name SYNC\
#            -space one_integer\
#            -intent OUT\
#            -comment {CGNS/Python sync integer out}
#
#PALM_OBJECT -name Dtable\
#            -space double_array\
#            -intent OUT\
#            -comment {CGNS/Python double array out}
#
#

# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
import sys
import numpy
import palm
from   mpi4py    import MPI
from   elsA.CGNS import core
import CGNS.MAP
from elsA_user import *
sys.path.append('/home/poinot/Tools/Dist/lib/py/elsA/Compat')
import elsAxdt  as E
import time

T=CGNS.MAP.load('SquaredNozzle-08-R_Inj1.hdf')

file_out='SquaredNozzle-result.hdf'
baseName='SquaredNozzle'
surfaceName='INJ1'
count=0

# start OpenPalm
palm.init('NONE')
local_communicator=MPI.Comm()
palm.get_mycomm(local_communicator)

# elsA interface initialized
sys.argv=['elsA_palm.py','-C','BndSubInj',
          '-P','elsA_exchange_script.py','-X','16',
          '--','elsA_main_script.py']
__e=core.elsA()
__e.initialize(sys.argv,local_communicator)
__e.parse(T)
__e.compute()
#__e.finalize()

# --- last line




# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
from elsA_user import *
import sys

sys.path.append('/home/poinot/Tools/Dist/lib/py/elsA/Compat')
import elsAxdt  as E
import time

elsAPB=cfdpb(name='elsAPB') ## For migration mode...

mytree=E.XdtCGNS("SquaredNozzle-08-R_Inj1.hdf")
mytree.action=E.TRANSLATE


(rank,size)=E.ranksize()
mytree.distribution={
    'SquaredNozzle/Zone-001':0,
    'SquaredNozzle/Zone-002':0,
    'SquaredNozzle/Zone-003':1,
    'SquaredNozzle/Zone-004':1,
    'SquaredNozzle/Zone-005':1,
    'SquaredNozzle/Zone-006':1,
    'SquaredNozzle/Zone-007':0,
    'SquaredNozzle/Zone-008':0
    }
mytree.compute()

##Migration mode##
R=sequence("blockName_%3d=mytree.e_getBlockInternalName('SquaredNozzle/Zone-%3d')", 1, 4)
R=sequence("blk_%3d=getBlock(blockName_%3d)", 1, 4)

#elsAPB.set('mpi_block2proc','procDict')

R=sequence("blk_%3d.show()", 1, 4)

## Extraction
R=sequence("ext_%3d=extractor(blockName_%3d,name='ext_%3d')", 1, 4)
R=sequence("ext_%3d.setDict({\
'var'        :'xyz conservative pgen',\
'file'       :'./Output/InletField_Inj1_blk%3d.tp',\
'format'     :'fmt_tp',\
'loc'        :'cellfict',\
'period'     : 50,\
'writingmode': 3,\
})", 1, 4)


elsAPB.compute()
elsAPB.extract() ##...ended migration mode

mytree.save("./Output/SquaredNozzle_OutMPI_#%03d.cgns" %rank)
#mytree.save("./Output/SquaredNozzle_CouplageNref.adf")





#PALM_UNIT   -name unit_Right\
#            -functions {python unit_Right}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_OBJECT -name Inode\
#            -space NULL\
#            -intent IN\
#            -comment {CGNS/Python double array in}
#
#PALM_OBJECT -name SYNC\
#            -space one_integer\
#            -intent IN\
#            -comment {CGNS/Python double array out}
#
import palm
from C3PO import capsule

palm.init('NONE')
palm.write('CAPSULE CGNS/Python')
v=capsule.nodeRetrieveAsDoubleArray("Inode")
palm.write('CoordinateX are is %s'%v)
palm.finalize()

# --- last line



#PALM_UNIT   -name unit_Right\
#            -functions {python unit_Right}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_OBJECT -name Inode\
#            -space NULL\
#            -intent IN\
#            -comment {CGNS/Python node data in}
#
import palm
from mpi4py import MPI
from C3PO import capsule

err = palm.init('NONE')
err = palm.write('CAPSULE CGNS/Python')

z=MPI.Comm()
palm.get_mycomm(z)
T=capsule.treeRetrieve("Inode")

err = palm.finalize()



#PALM_UNIT   -name fake_palm\
#            -functions {python fake_palm}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_OBJECT -name Dtable\
#            -space NULL\
#            -intent IN\
#            -comment {CGNS/Python double array in}
#
#PALM_OBJECT -name SYNC\
#            -space one_integer\
#            -intent IN\
#            -comment {CGNS/Python sync integer in}
#
#PALM_OBJECT -name Tnode\
#            -space NULL\
#            -intent IN\
#            -comment {CGNS/Python node data in}
#
import palm
import numpy 
import CGNS.MAP

T=CGNS.MAP.load('SquaredNozzle-08-R_Inj1.hdf')

baseName='SquaredNozzle'
surfaceName='INJ1'
count=0

palm.init('NONE')
local_communicator=MPI.Comm()
palm.get_mycomm(local_communicator)

local_data=numpy.ones((cpl.located,),dtype='d')*1.3
first_data=numpy.copy(local_data)
it=1

while (it<50):
  print "fake process: Couplage Stop iteration : %i" %it
  remote_data=cpl.publishAndRetrieve(local_data,iteration=it)
  local_data+=remote_data*.5
  if (not it%20):
    fwk.trace('fake total: %.3d=%e'%(it,local_data.sum()))
    local_delta=local_data-first_data
    fwk.attribute.local_delta=local_delta.max()
  it+=1
    
fwk.trace('leave fake')
del cpl

# --- last line



#PALM_UNIT   -name unit_Left\
#            -functions {python unit_Left}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_SPACE -name CGNSPYTHON\
#           -shape (:)\
#           -element_size PL_CHARACTER\
#           -comment {CGNS/Python sub-tree serialization}
#
#PALM_OBJECT -name Onode\
#            -space CGNSPYTHON\
#            -intent OUT\
#            -comment {CGNS/Python node data out}
#
import palm
from mpi4py import MPI
import T
from C3PO import capsule

err=palm.init('NONE')
err=palm.write('CAPSULE CGNS/Python')
z=MPI.Comm()
palm.get_mycomm(z) # to be used as elsA init input
capsule.treePublish("Onode","CGNSPYTHON.unit_Left",1,T.data)
err=palm.finalize()

# --- last line



#PALM_UNIT   -name unit_Right\
#            -functions {python unit_Right}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_OBJECT -name Inode\
#            -space one_integer\
#            -intent IN\
#            -comment {CGNS/Python single integer in}
#
import palm
from C3PO import capsule

palm.init('NONE')
palm.write('CAPSULE CGNS/Python')
v=capsule.nodeRetrieveAsInteger("Inode")
palm.write('niter is %f'%v)
palm.finalize()

# --- last line



# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------

# to be run with:
# export C3GENDOC
# C3GENDOC=1 sphinx-build -c doc -b html doc build/html 
# in the parent directory of this file

extensions = ['sphinx.ext.autodoc']
master_doc='index'
project = u'C3PO'
copyright = u'2013-, Marc Poinot'
version = '0'
release = '0.1'
unused_docs = ['license.txt']
source_suffix = '.txt'

pygments_style = 'sphinx'
html_title = "pyC3PO"
html_use_index = True

latex_paper_size = 'a4'
latex_font_size = '10pt'
latex_documents = [
  ("pyC3PO",
   'index.tex',
   u'pyC3PO - Coupling Capsule using CGNS/Python for OpenPalm Manual',
   u'Marc Poinot',
   'manual',False),
]
latex_use_parts = False
latex_use_modindex = True

autodoc_member_order='bysource'
autodoc_default_flags=['members', 'undoc-members']

# --- last line



# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
import sys

sys.path.append('/home/poinot/Tools/Dist/lib/py/elsA/Compat')

import elsAxdt as X
import numpy   as N
import copy    as C
from   math import *
from   C3PO import capsule

#############################################
def indg2D(i,j,Imax):
    """2D global indice located on nodes"""
    return i + ((j-1) * Imax)
#############################################

it     = X.iteration()
state  = X.state()
cname  = X.ccname()
(rank,size)=X.ranksize()

##for mpi : ######
if (rank == 0) :
    listZone=[1,2]
else :
    listZone=[3,4]
##################

if it>1:
    outputTree=C.deepcopy(X.get(X.OUTPUT_TREE))
else:
    outputTree=None

if (outputTree==[] or outputTree==None):
    outputTree= None ##[None,None,[],None]
elif(it>1) :
    dataNameList=['Pi','vx','vy','vz']
    ## Dictionary translating extracted variable with CGNS formulation ##
    parserDict={'Pi':'PressureStagnation','vx':'VelocityX','vy':'VelocityY','vz':'VelocityZ'}

    for iz in listZone:
        dataName='Pi'
        Path='/SquaredNozzle/Zone-%03i/ZoneBC/InflowSubsonic:West/BCDataSet/DirichletData/%s' %(iz,dataName)
        underTree=X.retrieve(Path,outputTree)
        NumArray=underTree[1]
        dim=NumArray.shape[0]
        ArraySend=N.ones([dim],'float')
        
        underTree[0]=parserDict[dataName]
        
        if (dataName == 'Pi') :
            for i in range(dim):
                ArraySend[i]=NumArray[i]+0.01
                
        if (iz==1):
          if (it>48):
              m=-1
          else:
              m=numpy.mean(ArraySend)
        
          capsule.nodePublishAsDoubleArray("Onode",'double_array.unit_Left',1,T.data,path)
          print "elsA process: Couplage Stop iteration : %i" %it
          local_data=numpy.ones((cpl.located,),dtype='d')*m
          remote_data=cpl.publishAndRetrieve(local_data,iteration=i)

        underTree[1]=ArraySend
        X.free(X.RUNTIME_TREE)
        X.free(X.OUTPUT_TREE)

    X.xdt(X.PYTHON,(X.RUNTIME_TREE,outputTree,1))    



#PALM_UNIT   -name unit_Left\
#            -functions {python unit_Left}\
#            -parallel mpi\
#            -minproc 1\
#            -maxproc 100000\
#            -comment {Capsule CGNS/Python}
#
#PALM_OBJECT -name Onode\
#            -space one_double\
#            -intent OUT\
#            -comment {CGNS/Python single double out}
#
import palm
import T
from C3PO import capsule

palm.init('NONE')
palm.write('CAPSULE CGNS/Python')
path='/CGNSTree/Cylindre/ReferenceState/Reynolds'
capsule.nodePublishAsDouble("Onode",1,T.data,path)
palm.write('Publish %s'%path)
palm.finalize()

# --- last line




# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
from setuptools import setup
from distutils.core import Extension
import sys


try:
    import palm
except ImportError:
    pass

# export C3POPALMINSTALL
# export C3POPALMINSTALL_MPI1
# export C3POPREPALMINSTALL
#
# C3POPALMINSTALL=/tmp_user/tiamat/poinot/COSMOS+/I2
# C3POPALMINSTALL_MPI1=/tmp_user/tiamat/poinot/COSMOS+/I1
# C3POPREPALMINSTALL=/tmp_user/tiamat/poinot/COSMOS+/PrePALM_MP
#
# python setup.py install --prefix=${C3POPALMINSTALL}
# (cd build/lib/C3PO/tests/t1; chmod 777 make.sh test.sh; make.sh; test.sh)
# (cd build/lib/C3PO/tests/t2; chmod 777 make.sh test.sh; make.sh; test.sh)
# ...
# python setup.py install --prefix=${C3POPALMINSTALL_MPI1}
# (cd build/lib/C3PO/tests/t6; chmod 777 make.sh test.sh; make.sh; test.sh)

# -------------------------------------------------------------------------
setup (
name         = "C3PO",
version      = '0.1',
description  = "Coupling Capsule using CGNS/Python for OpenPalm",
author       = "marc Poinot",
author_email = "marc.poinot@onera.fr",
license      = "LGPL 2",
packages     = ['C3PO'],
package_data = {'C3PO': \
                ['tests/t%d/*'%n for n in xrange(1,7)] +\
                ['tests/e%d/*'%n for n in xrange(1,2)] 
                },
)

# --- last line



# -------------------------------------------------------------------------
# C3PO - Coupling Capsule using CGNS/Python for OpenPalm
# See license.txt file in the root directory of this Python module source  
# -------------------------------------------------------------------------
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils as CGU
import CGNS.MAP
import numpy
import mpi4py.MPI as MPI

# ---
# this is a test case, all blocks have BC instead of BC & connectivities
# each block has 6 BCs, there is a base with only structured blocks and
# a base with unstructured blocks
# again, this is a test tree, this is not a usual CGNS pattern
#
SAVE=True

# ------------------------------------------------------------
imax=3
jmax=5
kmax=7

x=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
y=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
z=numpy.ones((imax,jmax,kmax),dtype='d',order='F')+2

for i in range(imax):
  for j in range(jmax):
    for k in range(kmax):
      x[i,j,k]=i
      y[i,j,k]=j
      z[i,j,k]=k

zsize=numpy.array([[imax,imax-1,0],[jmax,jmax-1,0],[kmax,kmax-1,0]],dtype='i',order='F')

T =CGL.newCGNSTree()
Tb=CGL.newCGNSBase(T,'Blocks:S',3,3)
Tf=CGL.newFamily(Tb,'ExchangeSurface')
Tu=CGL.newUserDefinedData(Tf,'DataInput')
Td=CGL.newDataArray(Tu,'Temperature')
Tu=CGL.newUserDefinedData(Tf,'DataOutput')
Td=CGL.newDataArray(Tu,'HeatFlux')
Tu=CGL.newUserDefinedData(Tf,'Mode')
Tf=CGL.newFamily(Tb,'ExchangeSurfaceStructured')
Tu=CGL.newUserDefinedData(Tf,'DataInput')
Td=CGL.newDataArray(Tu,'Temperature')
Tu=CGL.newUserDefinedData(Tf,'DataOutput')
Td=CGL.newDataArray(Tu,'HeatFlux')
Tu=CGL.newUserDefinedData(Tf,'Mode')
Tf=CGL.newFamily(Tb,'DOWN')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'EAST')
Tu=CGL.newFamilyBC(Tf,CGK.BCWallViscousIsothermal_s)
Tf=CGL.newFamily(Tb,'NORTH')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'WEST')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'SOUTH')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'UP')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)

# ------------------------------------------------------------
Tn=CGL.newZone(Tb,'Block:01:A',zsize)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,x/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Tq=CGL.newZoneBC(Tn)
T1=CGL.newBC(Tq,'D',numpy.array([[1,   imax],[1,   jmax],[1,   1   ]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='DOWN')
T2=CGL.newBC(Tq,'E',numpy.array([[imax,imax],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='EAST')
T3=CGL.newBC(Tq,'N',numpy.array([[1,   imax],[jmax,jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='NORTH')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T3)
T4=CGL.newBC(Tq,'W',numpy.array([[1,   1,  ],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='WEST')
T5=CGL.newBC(Tq,'S',numpy.array([[1,   imax],[1,   1   ],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='SOUTH')
T6=CGL.newBC(Tq,'U',numpy.array([[1,   imax],[1,   jmax],[kmax,kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='UP')
CGU.newNode('SurfaceOtherName',CGU.setStringAsArray('ExchangeSurfaceStructured'),[],CGK.AdditionalFamilyName_ts,T6)

# ------------------------------------------------------------
Tn=CGL.newZone(Tb,'Block:01:B',zsize)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,(x+imax-1)/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Tq=CGL.newZoneBC(Tn)
T1=CGL.newBC(Tq,'D',numpy.array([[1,   imax],[1,   jmax],[1,   1   ]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='DOWN')
T2=CGL.newBC(Tq,'E',numpy.array([[imax,imax],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='EAST')
T3=CGL.newBC(Tq,'N',numpy.array([[1,   imax],[jmax,jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='NORTH')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T3)
T4=CGL.newBC(Tq,'W',numpy.array([[1,   1,  ],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='WEST')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T4)
T5=CGL.newBC(Tq,'S',numpy.array([[1,   imax],[1,   1   ],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='SOUTH')
T6=CGL.newBC(Tq,'U',numpy.array([[1,   imax],[1,   jmax],[kmax,kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='UP')

# ------------------------------------------------------------
Tn=CGL.newZone(Tb,'Block:01:C',zsize)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,(x+2*(imax-1))/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Tq=CGL.newZoneBC(Tn)
T1=CGL.newBC(Tq,'D',numpy.array([[1,   imax],[1,   jmax],[1,   1   ]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='DOWN')
T2=CGL.newBC(Tq,'E',numpy.array([[imax,imax],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='EAST')
T3=CGL.newBC(Tq,'N',numpy.array([[1,   imax],[jmax,jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='NORTH')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T3)
T4=CGL.newBC(Tq,'W',numpy.array([[1,   1,  ],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='WEST')
T5=CGL.newBC(Tq,'S',numpy.array([[1,   imax],[1,   1   ],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='SOUTH')
T6=CGL.newBC(Tq,'U',numpy.array([[1,   imax],[1,   jmax],[kmax,kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='UP')

# ------------------------------------------------------------
imax=7
jmax=5
kmax=3

x=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
y=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
z=numpy.ones((imax,jmax,kmax),dtype='d',order='F')+2

for i in range(imax):
  for j in range(jmax):
    for k in range(kmax):
      x[i,j,k]=i
      y[i,j,k]=j
      z[i,j,k]=k+imax-1

zsize=numpy.array([[imax,imax-1,0],[jmax,jmax-1,0],[kmax,kmax-1,0]],dtype='i',order='F')

Tn=CGL.newZone(Tb,'Block:02:A',zsize)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,x/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Tq=CGL.newZoneBC(Tn)
T1=CGL.newBC(Tq,'D',numpy.array([[1,   imax],[1,   jmax],[1,   1   ]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='DOWN')
CGU.newNode('SurfaceOtherName',CGU.setStringAsArray('ExchangeSurfaceStructured'),[],CGK.AdditionalFamilyName_ts,T1)
T2=CGL.newBC(Tq,'E',numpy.array([[imax,imax],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='EAST')
T3=CGL.newBC(Tq,'N',numpy.array([[1,   imax],[jmax,jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='NORTH')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T3)
T4=CGL.newBC(Tq,'W',numpy.array([[1,   1,  ],[1,   jmax],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='WEST')
T5=CGL.newBC(Tq,'S',numpy.array([[1,   imax],[1,   1   ],[1,   kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='SOUTH')
T6=CGL.newBC(Tq,'U',numpy.array([[1,   imax],[1,   jmax],[kmax,kmax]],dtype='i',order='F'),btype=CGK.FamilySpecified_s,family='UP')

# ------------------------------------------------------------
x=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
y=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
z=numpy.ones((imax,jmax,kmax),dtype='d',order='F')

for i in range(imax):
  for j in range(jmax):
    for k in range(kmax):
      x[i,j,k]=i
      y[i,j,k]=j
      z[i,j,k]=k-kmax+1

Tn=CGL.newZone(Tb,'Block:03:A',zsize)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,x/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Tq=CGL.newZoneBC(Tn)
T1=CGL.newBC(Tq,'D',numpy.array([[1,   imax],[1,   jmax],[1,   1   ]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='DOWN')
T2=CGL.newBC(Tq,'E',numpy.array([[imax,imax],[1,   jmax],[1,   kmax]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='EAST')
T3=CGL.newBC(Tq,'N',numpy.array([[1,   imax],[jmax,jmax],[1,   kmax]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='NORTH')
CGU.newNode('SurfaceName',CGU.setStringAsArray('ExchangeSurface'),[],CGK.AdditionalFamilyName_ts,T3)
T4=CGL.newBC(Tq,'W',numpy.array([[1,   1,  ],[1,   jmax],[1,   kmax]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='WEST')
T5=CGL.newBC(Tq,'S',numpy.array([[1,   imax],[1,   1   ],[1,   kmax]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='SOUTH')
T6=CGL.newBC(Tq,'U',numpy.array([[1,   imax],[1,   jmax],[kmax,kmax]],
                                dtype='i',order='F'),
             btype=CGK.FamilySpecified_s,family='UP')

# ------------------------------------------------------------
imax=7
jmax=5
kmax=11

# add some random delta to coords in order to force interpolation
def dc():
    return (0.5-numpy.random.random())/10.

x=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
y=numpy.ones((imax,jmax,kmax),dtype='d',order='F')
z=numpy.ones((imax,jmax,kmax),dtype='d',order='F')+2

for i in range(imax):
  for j in range(jmax):
    for k in range(kmax):
      x[i,j,k]=i+dc()
      y[i,j,k]=j+jmax-1+dc()
      z[i,j,k]=k-2+dc()

# gets hexa index from structured
def topoIndex2D(imax,jmax,kmax,i,j,k,p=0):
  n1=(imax*jmax*(k+p))  +(imax*(j+p))  +i   +p+1
  n2=(imax*jmax*(k+p))  +(imax*(j+p))  +i+1 +p+1
  n3=(imax*jmax*(k+p))  +(imax*(j+1+p))+i+1 +p+1
  n4=(imax*jmax*(k+p))  +(imax*(j+1+p))+i   +p+1
  n5=(imax*jmax*(k+p+1))+(imax*(j+p))  +i   +p+1
  n6=(imax*jmax*(k+p+1))+(imax*(j+p))  +i+1 +p+1
  n7=(imax*jmax*(k+p+1))+(imax*(j+1+p))+i+1 +p+1
  n8=(imax*jmax*(k+p+1))+(imax*(j+1+p))+i   +p+1
  return (n1,n2,n3,n4,n5,n6,n7,n8)

hexa=[]
quad={'W':[],'E':[],'S':[],'N':[],'D':[],'U':[]}
n_hexas=0
n_quads=0
for i in range(imax-1):
  for j in range(jmax-1):
    for k in range(kmax-1):
      (n1,n2,n3,n4,n5,n6,n7,n8)=topoIndex2D(imax,jmax,kmax,i,j,k)
      hexa+=[n1,n2,n3,n4,n5,n6,n7,n8]
      n_hexas+=1
      if (i==0):
        quad['W']+=[n1,n5,n8,n4]
      if (i==imax-2):
        quad['E']+=[n2,n3,n7,n6]
      if (j==0):
        quad['S']+=[n1,n2,n6,n5]
      if (j==jmax-2):
        quad['N']+=[n3,n4,n8,n7]
      if (k==0):
        quad['D']+=[n1,n4,n3,n2]
      if (k==kmax-2):
        quad['U']+=[n5,n6,n7,n8]
      n_quads+=1

e=numpy.array(hexa,dtype='i')
n_points=imax*jmax*kmax
n_elems=n_hexas+n_quads

zsize=numpy.array([[n_points,n_elems,0]],dtype='i',order='F')
x=x.ravel(order='F')
y=y.ravel(order='F')
z=z.ravel(order='F')

Tb=CGL.newCGNSBase(T,'Blocks:U',3,3)
Tf=CGL.newFamily(Tb,'ExchangeSurface')
Tu=CGL.newUserDefinedData(Tf,'DataOutput')
Td=CGL.newDataArray(Tu,'Temperature')
Tu=CGL.newUserDefinedData(Tf,'DataIntput')
Td=CGL.newDataArray(Tu,'HeatFlux')
Tu=CGL.newUserDefinedData(Tf,'Mode')
Tf=CGL.newFamily(Tb,'DOWN')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'EAST')
Tu=CGL.newFamilyBC(Tf,CGK.BCWallViscousIsothermal_s)
Tf=CGL.newFamily(Tb,'NORTH')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'WEST')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'SOUTH')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tf=CGL.newFamily(Tb,'UP')
Tu=CGL.newFamilyBC(Tf,CGK.UserDefined_s)
Tn=CGL.newZone(Tb,'Block:04:A',zsize,ztype=CGK.Unstructured_s)
Tq=CGL.newZoneBC(Tn)
Tg=CGL.newGridCoordinates(Tn,CGK.GridCoordinates_s)
Tx=CGL.newDataArray(Tg,CGK.CoordinateX_s,x/11.)
Ty=CGL.newDataArray(Tg,CGK.CoordinateY_s,y/11.)
Tz=CGL.newDataArray(Tg,CGK.CoordinateZ_s,z/11.)
Te=CGL.newElements(Tn,'HEXA',CGK.HEXA_8,e,
                   erange=numpy.array([1,n_hexas],dtype='i'))

last_index=n_hexas+1
for n in ['WEST','EAST','NORTH','SOUTH','DOWN','UP']:
  e=numpy.array(quad[n[0]],dtype='i')
  n_quads=e.size/4
  erange=numpy.array([last_index,last_index+n_quads-1],dtype='i')
  last_index+=n_quads
  Te=CGL.newElements(Tn,'QUAD_%s'%n[0],CGK.QUAD_4_s,e,erange,eboundary=0)
  Tm=CGL.newBC(Tq,n[0],
               numpy.array([numpy.arange(erange[0],erange[1]+1)],dtype='i'),
               btype=CGK.FamilySpecified_s,
               pttype=CGK.PointList,
               family=n)
  if (n=='SOUTH'):
    CGU.newNode('SurfaceName',
                CGU.setStringAsArray('ExchangeSurface'),[],
                CGK.AdditionalFamilyName_ts,Tm)
  CGL.newGridLocation(Tm,CGK.FaceCenter_s)

bclist=[
  (1,   imax,1,   jmax,1,   1,   'DOWN'),
  (imax,imax,1,   jmax,1,   kmax,'EAST'),
  (1,   imax,jmax,jmax,1,   kmax,'NORTH'),
  (1,   1,   1,   jmax,1,   kmax,'WEST'),
  (1,   imax,1,   1,   1,   kmax,'SOUTH'),
  (1,   imax,1,   jmax,kmax,kmax,'UP')
  ]

if 0:
  last_index=n_hexas+1
  for (i1,i2,j1,j2,k1,k2,n) in bclist:
    quad=[]
    n_quads=0
    i2_=i2
    j2_=j2
    k2_=k2
    if (i1==i2): i2_=i2+1
    if (j1==j2): j2_=j2+1
    if (k1==k2): k2_=k2+1    
    for i in range(i1,i2_):
      for j in range(j1,j2_):
        for k in range(k1,k2_):
          (n1,n2,n3,n4,n5,n6,n7,n8)=topoIndex2D(imax,jmax,kmax,i-1,j-1,k-1)
          if ((i1==i2) and (i1==1)):    quad+=[n3,n4,n8,n7]
          if ((i1==i2) and (i1==imax)): quad+=[n1,n2,n6,n5]
          if ((j1==j2) and (j1==1)):    quad+=[n1,n5,n8,n4]
          if ((j1==j2) and (j1==jmax)): quad+=[n2,n3,n7,n6]
          if ((k1==k2) and (k1==1)):    quad+=[n1,n4,n3,n2]
          if ((k1==k2) and (k1==kmax)): quad+=[n5,n6,n7,n8]
          n_quads+=1
    e=numpy.array(quad,dtype='i')
    erange=numpy.array([last_index,last_index+n_quads-1],dtype='i')
    last_index+=n_quads
    Te=CGL.newElements(Tn,'QUAD_%s'%n[0],CGK.QUAD_4,e,erange,eboundary=0)
    Tm=CGL.newBC(Tq,n[0],
                 numpy.array([numpy.arange(erange[0],erange[1]+1)],dtype='i'),
                 btype=CGK.FamilySpecified_s,
                 pttype=CGK.PointList,
                 family=n)
    CGL.newGridLocation(Tm,CGK.FaceCenter_s)

# ------------------------------------------------------------
if (SAVE and not MPI.COMM_WORLD.rank):
  import os
  filename='blocks.hdf'
  if (os.path.exists(filename)): os.unlink(filename)
  CGNS.MAP.save(filename,T)

# --- last line
